// Simple ray tracer in a pixel shader
// - this is intended to demonstrate branching and early exit in PS 3.0
// - based on Paul Heckbert's code from "An Introduction to Ray Tracing"
// - it only supports spheres currently
// sgreen@nvidia.com 12/2003

/*
<scene_commands>
<clear depth="1.0" color = "0, 0,0,0"/>
<settechnique name="t0"/>
<draw type="buffer"/>
</scene_commands>
*/

float4x4 view : View;
float time : Time < string UIWidget = "none"; >;

float3 cop = { 0.0, 0.0, 0.0 };	// centre of projection
float2 viewport = { 100.0f, 100.0f };
float foclen <
string UIWidget="Slider";
string UIName="Focal Length";
float UIMin = 1.0f;
float UIMax = 500.0f;
float UIStep = 10.0f;
> = 1.0;	// focal length

float3 light_pos : POSITION
<
string UIName="Light Position";
string Space = "World";
> = { 10.0, -10.0, 10.0 };

float shininess
<
string UIWidget = "slider";
float UIMin = 1.0f;
float UIStep = 1.0f;
float UIMax = 100.0f;
> = 40.0;

float4 specularColor : SPECULAR
<
string UIName="Specular Color";
 string UIWidget = "color";
> = { 1.0, 1.0, 1.0, 1.0 };

struct Ray {
	float3 p;	// origin
	float3 r;	// direction
};

struct Sphere {
  float3 centre;
  float rad2;	// radius^2
  float4 color;
};

#define NOBJECTS 3
static Sphere object[NOBJECTS] = {
	{
		0.0, 0.0, 0.0,
		1.0,
		1.0, 0.0, 0.0, 1.0
	},
	{
		1.5, 0.5, 0.0,
		0.5*0.5,
		0.0, 1.0, 0.0, 1.0
	},	
	{
		0.0, 11.0, 0.0,
		10.0*10.0,
		0.0, 0.0, 1.0, 1.0
	}
};
const float eps = 0.001;	// error epsilon

float SphereIntersect(Sphere s, Ray ray, out bool hit)
{
  float3 v = s.centre - ray.p;
  float b = dot(v, ray.r);
  float disc = (b*b) - dot(v, v) + s.rad2;

  hit = true;
//  if (disc<=0) return -1.0; // ray misses
  if (disc<=0) hit = false;;

  disc = sqrt(disc);
  float t2 = b + disc;

//  if (t2<=eps) return -1.0; // behind ray origin
  if (t2<=eps) hit = false; // behind ray origin

  float t1 = b - disc;

  if ((t1>eps) && (t1<t2))  // return nearest intersection
    return t1;
  else
    return t2;
}

float3 SphereNormal(Sphere s, float3 i)
{
	return normalize(i - s.centre);
}

// find nearest hit
// returns intersection point
float3 NearestHit(Ray ray, out int hitobj)
{
	float mint = 1e10;
	hitobj = -1;
	for(int i=0; i<NOBJECTS; i++) {
		bool hit;
		float t = SphereIntersect(object[i], ray, hit);
		if (hit) {
			if (t < mint) {
				hitobj = i;
				mint = t;
			}
		}
	}
	return ray.p + ray.r*mint;
}

// test for any hit (for shadow rays)
bool AnyHit(Ray ray)
{
	bool anyhit = false;
	for(int i=0; i<NOBJECTS; i++) {
		bool hit;
		float t = SphereIntersect(object[i], ray, hit);
		if (hit) {
//			return true;
			anyhit = true;
		}
	}
	return anyhit;
}

// Phong lighting model
float4 Phong(float3 n, float3 l, float3 v, float shininess, float4 diffuseColor, float4 specularColor)
{
	float ndotl = dot(n, l);
	float diff = saturate(ndotl);
	float3 r = reflect(l, n);
	float spec = pow(dot(v, r), shininess) * (ndotl > 0.0);
	return diff*diffuseColor + spec*specularColor;
}

float4 Shade(float3 i, float3 n, float3 v, int hitobj)
{
	float3 l = normalize(light_pos - i);

	// check if shadowed
	Ray shadowray;
	shadowray.p = i;
	shadowray.r = l;
	bool shadowed = AnyHit(shadowray);

	// lighting
	return Phong(n, l, v, shininess, object[hitobj].color, specularColor) * !shadowed;
}

struct Vertex {
    float4 pos		: POSITION;
    float4 texcoord	: TEXCOORD0;
};

Vertex RayTraceVS(Vertex v)
{
    return v;
}

void animate_scene(float time)
{
	object[1].centre.x = sin(time)*1.5;
	object[1].centre.z = cos(time)*1.5;
}

float4 RayTracePS(Vertex IN) : COLOR
{
	animate_scene(time);
	
	// position on view plane
	float3 p;
	p.xy = ((IN.texcoord*2.0)-1.0) * viewport;
	p.z = cop.z - foclen;

	// calculate eye ray
	Ray eyeray;
	eyeray.p = cop;
	eyeray.r = normalize(p - cop);
	
	// transform eye ray
	eyeray.p = mul(float4(eyeray.p, 1.0), (float4x3) view);
	eyeray.r = mul(eyeray.r, (float3x3) view);
	
	// find nearest hit
	int hitobj;
	float3 i = NearestHit(eyeray, hitobj);
	float3 n = SphereNormal(object[hitobj], i);

	float4 c;
	c = Shade(i, n, eyeray.r, hitobj) * (hitobj > -1);

#if 1
	float facing = 1.0 - dot(-eyeray.r, n);
	float fresnel = 0.2 + 0.8 * pow(facing, 5.0);

	// shoot reflection ray
	float3 r = reflect(eyeray.r, n);
	Ray reflray;
	reflray.p = i;
	reflray.r = r;
	i = NearestHit(reflray, hitobj);
	n = SphereNormal(object[hitobj], i);
			
	c += Shade(i, n, reflray.r, hitobj) * (hitobj > -1);
//	c += Shade(i, n, reflray.r, hitobj) * fresnel * (hitobj > -1);
#endif

	return c;
//	return fresnel;
//	return hitobj>-1;
}

technique t0
{
    pass p0
    {
		VertexShader = compile vs_1_1 RayTraceVS();
		PixelShader = compile ps_2_a RayTracePS();
    }
}
